@@ -1,4 +1,4 @@
-# Copyright (c) 2016-2017 by Ron Frederick <ronf@timeheart.net>.else:
+# Copyright (c) 2016-2018 by Ron Frederick <ronf@timeheart.net>.else:
 # All rights reserved.
 #
 # This program and the accompanying materials are made available under
@@ -14,6 +14,7 @@
 
 import asyncio
 from asyncio.subprocess import DEVNULL, PIPE, STDOUT
+import codecs
 from collections import OrderedDict
 import io
 import os
@@ -39,42 +40,28 @@ def _is_regular_file(file):
 class _UnicodeReader:
     """Handle buffering partial Unicode data"""
 
-    def __init__(self, encoding, textmode=False):
-        self._encoding = encoding
-        self._textmode = textmode
-        self._partial = b''
+    def __init__(self, encoding, errors, textmode=False):
+        if encoding and not textmode:
+            self._decoder = codecs.getincrementaldecoder(encoding)(errors)
+        else:
+            self._decoder = None
 
-    def decode(self, data):
+    def decode(self, data, final=False):
         """Decode Unicode bytes when reading from binary sources"""
 
-        if self._encoding and not self._textmode:
-            data = self._partial + data
-            self._partial = b''
-
+        if self._decoder:
             try:
-                data = data.decode(self._encoding)
-            except UnicodeDecodeError as exc:
-                if exc.start > 0:
-                    # Avoid pylint false positive
-                    # pylint: disable=invalid-slice-index
-                    self._partial = data[exc.start:]
-                    data = data[:exc.start].decode(self._encoding)
-                elif exc.reason == 'unexpected end of data':
-                    self._partial = data
-                    data = ''
-                else:
-                    self.close()
-                    raise DisconnectError(DISC_PROTOCOL_ERROR,
-                                          'Unicode decode error')
+                data = self._decoder.decode(data, final)
+            except UnicodeDecodeError as unicode_exc:
+                raise DisconnectError(DISC_PROTOCOL_ERROR,
+                                      str(unicode_exc)) from None
 
         return data
 
     def check_partial(self):
         """Check if there's partial Unicode data left at EOF"""
 
-        if self._partial:
-            self.close()
-            raise DisconnectError(DISC_PROTOCOL_ERROR, 'Unicode decode error')
+        self.decode(b'', True)
 
     def close(self):
         """Perform necessary cleanup on error (provided by derived classes)"""
@@ -83,15 +70,17 @@ class _UnicodeReader:
 class _UnicodeWriter:
     """Handle encoding Unicode data before writing it"""
 
-    def __init__(self, encoding, textmode=False):
-        self._encoding = encoding
-        self._textmode = textmode
+    def __init__(self, encoding, errors, textmode=False):
+        if encoding and not textmode:
+            self._encoder = codecs.getincrementalencoder(encoding)(errors)
+        else:
+            self._encoder = None
 
     def encode(self, data):
         """Encode Unicode bytes when writing to binary targets"""
 
-        if self._encoding and not self._textmode:
-            data = data.encode(self._encoding)
+        if self._encoder:
+            data = self._encoder.encode(data)
 
         return data
 
@@ -99,8 +88,8 @@ class _UnicodeWriter:
 class _FileReader(_UnicodeReader):
     """Forward data from a file"""
 
-    def __init__(self, process, file, bufsize, datatype, encoding):
-        super().__init__(encoding, hasattr(file, 'encoding'))
+    def __init__(self, process, file, bufsize, datatype, encoding, errors):
+        super().__init__(encoding, errors, hasattr(file, 'encoding'))
 
         self._process = process
         self._file = file
@@ -141,8 +130,8 @@ class _FileReader(_UnicodeReader):
 class _AsyncFileReader(_FileReader):
     """Forward data from an aiofile"""
 
-    def __init__(self, process, file, bufsize, datatype, encoding):
-        super().__init__(process, file, bufsize, datatype, encoding)
+    def __init__(self, process, file, bufsize, datatype, encoding, errors):
+        super().__init__(process, file, bufsize, datatype, encoding, errors)
 
         self._conn = process.channel.get_connection()
 
@@ -174,8 +163,8 @@ class _AsyncFileReader(_FileReader):
 class _FileWriter(_UnicodeWriter):
     """Forward data to a file"""
 
-    def __init__(self, file, encoding):
-        super().__init__(encoding, hasattr(file, 'encoding'))
+    def __init__(self, file, encoding, errors):
+        super().__init__(encoding, errors, hasattr(file, 'encoding'))
 
         self._file = file
 
@@ -198,8 +187,8 @@ class _FileWriter(_UnicodeWriter):
 class _AsyncFileWriter(_FileWriter):
     """Forward data to an aiofile"""
 
-    def __init__(self, process, file, encoding):
-        super().__init__(file, encoding)
+    def __init__(self, process, file, encoding, errors):
+        super().__init__(file, encoding, errors)
 
         self._conn = process.channel.get_connection()
 
@@ -217,8 +206,8 @@ class _AsyncFileWriter(_FileWriter):
 class _PipeReader(_UnicodeReader, asyncio.Protocol):
     """Forward data from a pipe"""
 
-    def __init__(self, process, datatype, encoding):
-        super().__init__(encoding)
+    def __init__(self, process, datatype, encoding, errors):
+        super().__init__(encoding, errors)
 
         self._process = process
         self._datatype = datatype
@@ -259,8 +248,8 @@ class _PipeReader(_UnicodeReader, asyncio.Protocol):
 class _PipeWriter(_UnicodeWriter, asyncio.BaseProtocol):
     """Forward data to a pipe"""
 
-    def __init__(self, process, datatype, encoding):
-        super().__init__(encoding)
+    def __init__(self, process, datatype, encoding, errors):
+        super().__init__(encoding, errors)
 
         self._process = process
         self._datatype = datatype
@@ -574,7 +563,7 @@ class SSHProcess:
         def pipe_factory():
             """Return a pipe read handler"""
 
-            return _PipeReader(self, datatype, self._encoding)
+            return _PipeReader(self, datatype, self._encoding, self._errors)
 
         if source == PIPE:
             reader = None
@@ -599,11 +588,11 @@ class SSHProcess:
                 file = source
 
             if hasattr(file, 'read') and asyncio.iscoroutinefunction(file.read):
-                reader = _AsyncFileReader(self, file, bufsize,
-                                          datatype, self._encoding)
+                reader = _AsyncFileReader(self, file, bufsize, datatype,
+                                          self._encoding, self._errors)
             elif _is_regular_file(file):
-                reader = _FileReader(self, file, bufsize,
-                                     datatype, self._encoding)
+                reader = _FileReader(self, file, bufsize, datatype,
+                                     self._encoding, self._errors)
             else:
                 if hasattr(source, 'buffer'):
                     # If file was opened in text mode, remove that wrapper
@@ -626,7 +615,7 @@ class SSHProcess:
         def pipe_factory():
             """Return a pipe write handler"""
 
-            return _PipeWriter(self, datatype, self._encoding)
+            return _PipeWriter(self, datatype, self._encoding, self._errors)
 
         if target == DEVNULL:
             writer = _DevNullWriter()
@@ -653,9 +642,10 @@ class SSHProcess:
 
             if hasattr(file, 'write') and \
                     asyncio.iscoroutinefunction(file.write):
-                writer = _AsyncFileWriter(self, file, self._encoding)
+                writer = _AsyncFileWriter(self, file, self._encoding,
+                                          self._errors)
             elif _is_regular_file(file):
-                writer = _FileWriter(file, self._encoding)
+                writer = _FileWriter(file, self._encoding, self._errors)
             else:
                 if hasattr(target, 'buffer'):
                     # If file was opened in text mode, remove that wrapper
