@@ -1,4 +1,4 @@
-# Copyright (c) 2013-2015 by Ron Frederick <ronf@timeheart.net>.
+# Copyright (c) 2013-2017 by Ron Frederick <ronf@timeheart.net>.
 # All rights reserved.
 #
 # This program and the accompanying materials are made available under
@@ -15,10 +15,10 @@
 from .asn1 import ASN1DecodeError, ObjectIdentifier, der_encode, der_decode
 from .crypto import RSAPrivateKey, RSAPublicKey
 from .misc import all_ints
-from .packet import MPInt, String, PacketDecodeError, SSHPacket
-from .public_key import SSHKey, SSHCertificateV01
-from .public_key import KeyExportError
+from .packet import MPInt
+from .public_key import SSHKey, SSHOpenSSHCertificateV01, KeyExportError
 from .public_key import register_public_key_alg, register_certificate_alg
+from .public_key import register_x509_certificate_alg
 
 # Short variable names are used here, matching names in the spec
 # pylint: disable=invalid-name
@@ -30,9 +30,10 @@ class _RSAKey(SSHKey):
     algorithm = b'ssh-rsa'
     pem_name = b'RSA'
     pkcs8_oid = ObjectIdentifier('1.2.840.113549.1.1.1')
-
-    def __init__(self, key):
-        self._key = key
+    sig_algorithms = (b'rsa-sha2-256', b'rsa-sha2-512', b'ssh-rsa')
+    x509_sig_algorithms = (b'rsa2048-sha256', b'ssh-rsa')
+    x509_algorithms = tuple(b'x509v3-' + alg for alg in x509_sig_algorithms)
+    all_sig_algorithms = set(x509_sig_algorithms + sig_algorithms)
 
     def __eq__(self, other):
         # This isn't protected access - both objects are _RSAKey instances
@@ -56,16 +57,16 @@ class _RSAKey(SSHKey):
         return cls(RSAPrivateKey.generate(key_size, exponent))
 
     @classmethod
-    def make_private(cls, *args):
+    def make_private(cls, n, e, d, p, q, dmp1, dmq1, iqmp):
         """Construct an RSA private key"""
 
-        return cls(RSAPrivateKey.construct(*args))
+        return cls(RSAPrivateKey.construct(n, e, d, p, q, dmp1, dmq1, iqmp))
 
     @classmethod
-    def make_public(cls, *args):
+    def make_public(cls, n, e):
         """Construct an RSA public key"""
 
-        return cls(RSAPublicKey.construct(*args))
+        return cls(RSAPublicKey.construct(n, e))
 
     @classmethod
     def decode_pkcs1_private(cls, key_data):
@@ -176,33 +177,57 @@ class _RSAKey(SSHKey):
 
         return b''.join((MPInt(self._key.e), MPInt(self._key.n)))
 
-    def sign(self, data):
-        """Return a signature of the specified data using this key"""
+    def encode_agent_cert_private(self):
+        """Encode RSA certificate private key data for agent"""
+
+        if not self._key.d:
+            raise KeyExportError('Key is not private')
+
+        return b''.join((MPInt(self._key.d), MPInt(self._key.iqmp),
+                         MPInt(self._key.p), MPInt(self._key.q)))
+
+    def sign_der(self, data, sig_algorithm):
+        """Compute a DER-encoded signature of the specified data"""
+
+        # pylint: disable=unused-argument
 
         if not self._key.d:
             raise ValueError('Private key needed for signing')
 
-        sig = self._key.sign(data)
-        return b''.join((String(self.algorithm), String(sig)))
+        return self._key.sign(data, sig_algorithm)
 
-    def verify(self, data, sig):
-        """Verify a signature of the specified data using this key"""
+    def verify_der(self, data, sig_algorithm, sig):
+        """Verify a DER-encoded signature of the specified data"""
 
-        try:
-            packet = SSHPacket(sig)
+        # pylint: disable=unused-argument
 
-            if packet.get_string() != self.algorithm:
-                return False
+        return self._key.verify(data, sig, sig_algorithm)
 
-            sig = packet.get_string()
-            packet.check_end()
+    def sign_ssh(self, data, sig_algorithm):
+        """Compute an SSH-encoded signature of the specified data"""
 
-            return self._key.verify(data, sig)
-        except PacketDecodeError:
-            return False
+        return self.sign_der(data, sig_algorithm)
+
+    def verify_ssh(self, data, sig_algorithm, sig):
+        """Verify an SSH-encoded signature of the specified data"""
+
+        return self.verify_der(data, sig_algorithm, sig)
+
+    def encrypt(self, data, algorithm):
+        """Encrypt a block of data with this key"""
+
+        return self._key.encrypt(data, algorithm)
+
+    def decrypt(self, data, algorithm):
+        """Decrypt a block of data with this key"""
+
+        return self._key.decrypt(data, algorithm)
 
 
 register_public_key_alg(b'ssh-rsa', _RSAKey)
 
-register_certificate_alg(1, b'ssh-rsa-cert-v01@openssh.com',
-                         _RSAKey, SSHCertificateV01)
+register_certificate_alg(1, b'ssh-rsa', b'ssh-rsa-cert-v01@openssh.com',
+                         _RSAKey, SSHOpenSSHCertificateV01)
+
+for alg in _RSAKey.x509_algorithms:
+    register_x509_certificate_alg(alg)
