@@ -14,6 +14,7 @@
 
 import asyncio
 import binascii
+import codecs
 
 from .constants import DEFAULT_LANG, DISC_PROTOCOL_ERROR, EXTENDED_DATA_STDERR
 from .constants import MSG_CHANNEL_OPEN, MSG_CHANNEL_WINDOW_ADJUST
@@ -45,7 +46,7 @@ class SSHChannel(SSHPacketHandler):
     _read_datatypes = set()
     _write_datatypes = set()
 
-    def __init__(self, conn, loop, encoding, window, max_pktsize):
+    def __init__(self, conn, loop, encoding, errors, window, max_pktsize):
         """Initialize an SSH channel
 
            If encoding is set, data sent and received will be in the form
@@ -62,7 +63,6 @@ class SSHChannel(SSHPacketHandler):
         self._conn = conn
         self._loop = loop
         self._session = None
-        self._encoding = encoding
         self._extra = {'connection': conn}
 
         self._env = {}
@@ -81,9 +81,8 @@ class SSHChannel(SSHPacketHandler):
         self._init_recv_window = window
         self._recv_window = window
         self._recv_pktsize = max_pktsize
-        self._recv_paused = True
+        self._recv_paused = 'starting'
         self._recv_buf = []
-        self._recv_partial = {}
 
         self._request_queue = []
 
@@ -97,6 +96,7 @@ class SSHChannel(SSHPacketHandler):
         self._logger = conn.logger.get_child(context='chan=%d' %
                                              self._recv_chan)
 
+        self.set_encoding(encoding, errors)
         self.set_write_buffer_limits()
 
     @property
@@ -118,12 +118,20 @@ class SSHChannel(SSHPacketHandler):
     def get_encoding(self):
         """Return the encoding used by this channel"""
 
-        return self._encoding
+        return self._encoding, self._errors
 
-    def set_encoding(self, encoding):
+    def set_encoding(self, encoding, errors='strict'):
         """Set the encoding on this channel"""
 
         self._encoding = encoding
+        self._errors = errors
+
+        if encoding:
+            self._encoder = codecs.getincrementalencoder(encoding)(errors)
+            self._decoder = codecs.getincrementaldecoder(encoding)(errors)
+        else:
+            self._encoder = None
+            self._decoder = None
 
     def get_recv_window(self):
         """Return the configured receive window for this channel"""
@@ -171,7 +179,10 @@ class SSHChannel(SSHPacketHandler):
             self.logger.info('Channel closed%s',
                              ': ' + str(exc) if exc else '')
 
+            self._conn.detach_x11_listener(self)
+
             self._conn.remove_channel(self._recv_chan)
+            self._send_chan = None
             self._recv_chan = None
             self._conn = None
 
@@ -199,6 +210,18 @@ class SSHChannel(SSHPacketHandler):
             self._recv_state = 'closed'
             self._loop.call_soon(self._cleanup)
 
+    @asyncio.coroutine
+    def _start_reading(self):
+        """Start processing data on a new connection"""
+
+        # If owner of the channel  didn't explicitly pause it at
+        # startup, begin processing incoming data.
+
+        if self._recv_paused == 'starting':
+            self.logger.debug2('Reading from channel started')
+            self._recv_paused = False
+            self._flush_recv_buf()
+
     def _pause_resume_writing(self):
         """Pause or resume writing based on send buffer low/high water marks"""
 
@@ -254,6 +277,14 @@ class SSHChannel(SSHPacketHandler):
             self._deliver_data(*self._recv_buf.pop(0))
 
         if not self._recv_buf:
+            if self._encoding and not exc and \
+                    self._recv_state in ('eof_pending', 'close_pending'):
+                try:
+                    self._decoder.decode(b'', True)
+                except UnicodeDecodeError as unicode_exc:
+                    raise DisconnectError(DISC_PROTOCOL_ERROR,
+                                          str(unicode_exc)) from None
+
             if self._recv_state == 'eof_pending':
                 self._recv_state = 'eof'
 
@@ -262,11 +293,6 @@ class SSHChannel(SSHPacketHandler):
                     self.write_eof()
             elif self._recv_state == 'close_pending':
                 self._recv_state = 'closed'
-
-                if self._recv_partial and not exc:
-                    exc = DisconnectError(DISC_PROTOCOL_ERROR,
-                                          'Unicode decode error')
-
                 self._loop.call_soon(self._cleanup, exc)
 
     def _deliver_data(self, data, datatype):
@@ -284,32 +310,13 @@ class SSHChannel(SSHPacketHandler):
             self._recv_window = self._init_recv_window
 
         if self._encoding:
-            if datatype in self._recv_partial:
-                encdata = self._recv_partial.pop(datatype) + data
-            else:
-                encdata = data
+            try:
+                data = self._decoder.decode(data)
+            except UnicodeDecodeError as unicode_exc:
+                raise DisconnectError(DISC_PROTOCOL_ERROR,
+                                      str(unicode_exc)) from None
 
-            while encdata:
-                try:
-                    data = encdata.decode(self._encoding)
-                    encdata = b''
-                except UnicodeDecodeError as exc:
-                    if exc.start > 0:
-                        # Avoid pylint false positive
-                        # pylint: disable=invalid-slice-index
-                        data = encdata[:exc.start].decode()
-                        encdata = encdata[exc.start:]
-                    elif exc.reason == 'unexpected end of data':
-                        break
-                    else:
-                        raise DisconnectError(DISC_PROTOCOL_ERROR,
-                                              'Unicode decode error')
-
-                self._session.data_received(data, datatype)
-
-            if encdata:
-                self._recv_partial[datatype] = encdata
-        else:
+        if self._session:
             self._session.data_received(data, datatype)
 
     def _accept_data(self, data, datatype=None):
@@ -388,14 +395,7 @@ class SSHChannel(SSHPacketHandler):
 
         self.logger.info('Closing channel due to connection close')
 
-        if self._send_state != 'closed':
-            self._send_state = 'closed'
-
-        if self._recv_state not in {'close_pending', 'closed'}:
-            self._recv_state = 'close_pending'
-            self._flush_recv_buf(exc)
-        elif self._recv_state == 'closed':
-            self._loop.call_soon(self._cleanup, exc)
+        self._cleanup(exc)
 
     def process_open(self, send_chan, send_window, send_pktsize, session):
         """Process a channel open request"""
@@ -424,6 +424,10 @@ class SSHChannel(SSHPacketHandler):
             if asyncio.iscoroutine(session):
                 session = yield from session
 
+            if not self._conn:
+                raise ChannelOpenError(OPEN_CONNECT_FAILED,
+                                       'SSH connection closed')
+
             chan, self._session = self._wrap_session(session)
 
             self.logger.debug2('  Initial recv window %d, packet size %d',
@@ -439,8 +443,10 @@ class SSHChannel(SSHPacketHandler):
 
             self._session.connection_made(chan)
         except ChannelOpenError as exc:
-            self._conn.send_channel_open_failure(self._send_chan, exc.code,
-                                                 exc.reason, exc.lang)
+            if self._conn:
+                self._conn.send_channel_open_failure(self._send_chan, exc.code,
+                                                     exc.reason, exc.lang)
+
             self._loop.call_soon(self._cleanup)
 
     def process_open_confirmation(self, send_chan, send_window,
@@ -626,8 +632,7 @@ class SSHChannel(SSHPacketHandler):
         self._conn.send_packet(MSG_CHANNEL_OPEN, String(chantype),
                                UInt32(self._recv_chan),
                                UInt32(self._recv_window),
-                               UInt32(self._recv_pktsize), *args,
-                               handler_names=self._handler_names)
+                               UInt32(self._recv_pktsize), *args, handler=self)
 
         return (yield from self._open_waiter)
 
@@ -639,8 +644,7 @@ class SSHChannel(SSHPacketHandler):
 
         payload = UInt32(self._send_chan) + b''.join(args)
 
-        self._conn.send_packet(pkttype, payload,
-                               handler_names=self._handler_names)
+        self._conn.send_packet(pkttype, payload, handler=self)
 
     def _send_request(self, request, *args, want_reply=False):
         """Send a channel request"""
@@ -825,7 +829,7 @@ class SSHChannel(SSHPacketHandler):
             return
 
         if self._encoding:
-            data = data.encode(self._encoding)
+            data = self._encoder.encode(data)
 
         if datatype:
             typename = ' to %s' % _data_type_names[datatype]
@@ -977,20 +981,12 @@ class SSHClientChannel(SSHChannel):
 
     _read_datatypes = {EXTENDED_DATA_STDERR}
 
-    def __init__(self, conn, loop, encoding, window, max_pktsize):
-        super().__init__(conn, loop, encoding, window, max_pktsize)
+    def __init__(self, conn, loop, encoding, errors, window, max_pktsize):
+        super().__init__(conn, loop, encoding, errors, window, max_pktsize)
 
         self._exit_status = None
         self._exit_signal = None
 
-    def _cleanup(self, exc=None):
-        """Clean up this channel"""
-
-        if self._conn: # pragma: no branch
-            self._conn.detach_x11_listener(self)
-
-        super()._cleanup(exc)
-
     @asyncio.coroutine
     def create(self, session_factory, command, subsystem, env, term_type,
                term_size, term_modes, x11_forwarding, x11_display,
@@ -1070,7 +1066,9 @@ class SSHClientChannel(SSHChannel):
                 String(binascii.b2a_hex(remote_auth)), UInt32(screen))
 
             if not result:
-                self._conn.detach_x11_listener(self)
+                if self._conn: # pragma: no branch
+                    self._conn.detach_x11_listener(self)
+
                 raise ChannelOpenError(OPEN_REQUEST_X11_FORWARDING_FAILED,
                                        'X11 forwarding request failed')
 
@@ -1095,9 +1093,9 @@ class SSHClientChannel(SSHChannel):
                                    'Session request failed')
 
         self._session.session_started()
-        self.resume_reading()
+        self._conn.create_task(self._start_reading(), self.logger)
 
-        return self, self._session
+        return self._session
 
     def _process_xon_xoff_request(self, packet):
         """Process a request to set up XON/XOFF processing"""
@@ -1265,6 +1263,11 @@ class SSHClientChannel(SSHChannel):
 
            :raises: :exc:`OSError` if the channel is not open
 
+           .. note:: If your server-side runs on OpenSSH,
+                     this might be ineffective;
+                     for more details, see the note in
+                     :meth:`send_signal`
+
         """
 
         self.send_signal('TERM')
@@ -1277,6 +1280,11 @@ class SSHClientChannel(SSHChannel):
 
            :raises: :exc:`OSError` if the channel is not open
 
+           .. note:: If your server-side runs on OpenSSH,
+                     this might be ineffective;
+                     for more details, see the note in
+                     :meth:`send_signal`
+
         """
 
         self.send_signal('KILL')
@@ -1288,10 +1296,10 @@ class SSHServerChannel(SSHChannel):
     _write_datatypes = {EXTENDED_DATA_STDERR}
 
     def __init__(self, conn, loop, allow_pty, line_editor, line_history,
-                 encoding, window, max_pktsize):
+                 encoding, errors, window, max_pktsize):
         """Initialize an SSH server channel"""
 
-        super().__init__(conn, loop, encoding, window, max_pktsize)
+        super().__init__(conn, loop, encoding, errors, window, max_pktsize)
 
         self._env = conn.get_key_option('environment', {})
 
@@ -1305,13 +1313,6 @@ class SSHServerChannel(SSHChannel):
 
         self.logger.info('New SSH session requested')
 
-    def _cleanup(self, exc=None):
-        """Clean up this channel"""
-
-        self._conn.detach_x11_listener(self)
-
-        super()._cleanup(exc)
-
     def _wrap_session(self, session):
         """Wrap a line editor around the session if enabled"""
 
@@ -1409,7 +1410,6 @@ class SSHServerChannel(SSHChannel):
         self._conn.create_task(self._finish_x11_req_request(auth_proto,
                                                             auth_data, screen),
                                self.logger)
-        return None
 
     @asyncio.coroutine
     def _finish_x11_req_request(self, auth_proto, auth_data, screen):
@@ -1534,7 +1534,7 @@ class SSHServerChannel(SSHChannel):
             self.logger.info('Received window change: %sx%s (%sx%s pixels)',
                              width, height, pixwidth, pixheight)
         else:
-            self.logger.info('Recceived indow change: %sx%s', width, height)
+            self.logger.info('Recceived window change: %sx%s', width, height)
 
         self._term_size = (width, height, pixwidth, pixheight)
         self._session.terminal_size_changed(width, height, pixwidth, pixheight)
@@ -1805,9 +1805,8 @@ class SSHForwardChannel(SSHChannel):
 
         self._session = session_factory()
         self._session.connection_made(self)
-
         self._session.session_started()
-        self.resume_reading()
+        self._conn.create_task(self._start_reading(), self.logger)
 
         return self._session
 
